/*
Union/Find Algorithm Visualization:  Part of the algoviz@vt collection of algorithm visualizations.
Copyright (C) 2004, 2007  Cris Kania, Cliff Shaffer

This file is part of the Union/Find Algorithm Visualization.

Union/Find Algorithm Visualization is free software: you can redistribute
it and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

Union/Find Algorithm Visualization is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along with
Union/Find Algorithm Visualization.  If not, see
<http://www.gnu.org/licenses/>.
*/
import java.util.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.text.*;

public class ParPtrTree {
	// colors to use for selected nodes
	final Color nColor1 = Color.BLUE;
	final Color nColor2 = new Color(0, 192, 0);
	final Color selColor = Color.YELLOW;

	// nodes clicked on for the union
	int unode1, unode2, root1, root2;

	// insert for compression
	int ins_node;
	boolean first;

	// record messages here
	Document doc;
	JTextPane textPane;
    
	// must remeber path to root for path compression
	Vector<Integer> path;

	// with compression on, the tree is simplified on each
	// call to FIND (path compression)
	boolean compressionOn;

	// for stepping through the algorithm
	int stepper;

	// Node maintains index of parent and its data value
	// for graphic visualization, it stores point location,
	// width and a list of its children
	public class Node {
		int parent;
		char value;
		int weight;
		// selected under path from one of the 2 nodes
		boolean path;
		// prev, next location; interpolated over time
		Point p, p0, p1;
		// pixels that the children span
		int width, height;
		// refs of nodes
		Vector<Node> children;
		// set up default values
		public Node() {
			parent = -1;
			weight = 1;
			value = '?';
			path = false;
			p = new Point();
			p0 = new Point();
			p1 = new Point();
			width = 0;
			height = 0;
			children = new Vector<Node>();
		}
		public boolean isAlone()
		{
			return parent == -1 && children.size() == 0;
		}
	}

	// list of nodes in the container
	protected Node[] array;
	// refs of nodes
	protected Vector<Node> roots;

	// retrieve int value of last path element
	protected int lpath() {
		return ((Integer)path.get(path.size()-1)).intValue();
	}

	// traverse down to the root of this node or path
	// keeps track of path, return true if at root
	protected boolean FIND_down(int i) {
		if (path.size() == 0)
		{
			path.add(new Integer(i));
			array[i].path = true;
		}
		int ln = ((Integer)path.get(path.size()-1)).intValue();
		if (array[ln].parent == -1) {
			if (stepper == 0)
			{
				root1 = ln;
				output("The root of " + array[unode1].value + " is " + array[root1].value + "\n", 2);
			}
			else if (stepper == 2)
			{
				root2 = ln;
				output("The root of " + array[unode2].value + " is " + array[root2].value + "\n", 2);
			}
			return true;
		}
		array[array[ln].parent].path = true;
		path.add(new Integer(array[ln].parent));
		return false;
	}

	// return up the tree, compressing if necessary
	// clean up path variable
	protected boolean FIND_up() {
		if (path.size() == 0) return true;
		int ln = ((Integer)path.remove(path.size()-1)).intValue();
		array[ln].path = false;
		if (compressionOn) {
			output("Making " + array[(stepper == 1) ? root1 : root2].value + " the parent of " + array[ln].value + "\n", 2);
			if (array[ln].parent == ((stepper == 1) ? root1 : root2)) return (path.size() == 0);
			array[array[ln].parent].children.remove(array[ln]);
			array[ln].parent = (stepper == 1) ? root1 : root2;
			array[array[ln].parent].children.insertElementAt(array[ln], ins_node);
			ins_node++;
			UFVisual.remake = true;
		}
		return (path.size() == 0);
	}
	
	// create a tree given its node capacity
	public ParPtrTree(int sz) 
	{
		unode1 = unode2 = root1 = root2 = -1;
		path = new Vector<Integer>();
		stepper = -1;
		first = true;
		reset(sz);
	}

	// step the simulation; return true if more steps are to follow
	public boolean step() {
		switch (stepper) {
			case 0:
				if (first)
				{
					output("Finding root of " + array[unode1].value + "\n", 1);
					first = false;
				}
				if (FIND_down(unode1)) {
					first = true;
					if (compressionOn)
					{
						stepper++;
						if (path.size() > 2) ins_node = array[root1].children.indexOf(array[((Integer)path.get(path.size()-2)).intValue()])+1;
						int ln = ((Integer)path.remove(path.size()-1)).intValue();
						array[ln].path = false;
					}
					else 
					{
						for (int i = 0; i < path.size(); i++) array[((Integer)path.get(i)).intValue()].path = false;
						path.clear();
						stepper += 2;
					}
				}
				break;
			case 1:
				if (first)
				{
					output("Compress path from " + array[root1].value + " to " + array[unode1].value + "\n", 2);
					first = false;
				}
				if (FIND_up())
				{
					first = true;
					stepper++;
				}
				break;
			case 2:
				if (first)
				{
					output("Finding root of " + array[unode2].value + "\n", 1);
					first = false;
				}
				if (FIND_down(unode2)) 
				{
					first = true;
					if (compressionOn)
					{
						stepper++;
						if (path.size() > 2) ins_node = array[root2].children.indexOf(array[((Integer)path.get(path.size()-2)).intValue()])+1;
						int ln = ((Integer)path.remove(path.size()-1)).intValue();
						array[ln].path = false;
					}
					else 
					{
						for (int i = 0; i < path.size(); i++) array[((Integer)path.get(i)).intValue()].path = false;
						path.clear();
						stepper += 2;
					}
				}
				break;
			case 3:
				if (first)
				{
					output("Compress path from " + array[root2].value + " to " + array[unode2].value + "\n", 2);
					first = false;
				}
				if (FIND_up())
				{
					first = true;
					stepper++;
				}
				break;
			case 4:
				if (root1 != root2) 
				{
					if (array[root1].weight >= array[root2].weight)
					{
						output("Making " + array[root1].value + " the parent of " + array[root2].value + "\n", 1);
						array[root2].parent = root1;
						array[root1].weight += array[root2].weight;
						// handle the graphics mess
						array[root1].children.add(array[root2]);
						roots.remove(array[root2]);
					}
					else
					{
						output("Making " + array[root2].value + " the parent of " + array[root1].value + "\n", 1);
						array[root1].parent = root2;
						array[root2].weight += array[root1].weight;
						// handle the graphics mess
						array[root2].children.add(array[root1]);
						roots.remove(array[root1]);
					}
					UFVisual.remake = true;
					stepper++;
				} 
				else 
				{
					output("Nodes " + array[unode1].value + " and " + array[unode2].value + " are already unioned\n", 1);
					stepper += 2;
					unode1 = unode2 = root1 = root2 = -1;
				}
				break;
			case 5:
				stepper++;
				unode1 = unode2 = root1 = root2 = -1;
				break;
		}
		if (stepper == 6) 
		{
			stepper = -1;
			return true;
		} else return false;
	}

	// allocate a new list of nodes for the container
	public void reset(int sz) {
		// make sure # of nodes <= 26
		if (sz > 26) sz = 26;

		// allocate the array of nodes
		array = new Node[sz];

		// set up a root node to keep track of parents
		roots = new Vector<Node>(sz);

		path.clear();
		unode1 = unode2 = root1 = root2 = -1;
		first = true;
		stepper = -1;

		// remove random values from the set and place in tree
		for (int i = 0; i < array.length; ++i) {
			array[i] = new Node();
			array[i].value = (char)(i + 'A');
			roots.add(array[i]);
		}

		UFVisual.remake = true;
	}

	// form a union between two nodes
	public void UNION(int m, int n) {
		output("UNION " + array[m].value + " and " + array[n].value + "\n", 0);
		unode1 = m;
		unode2 = n;
		stepper = 0;
		root1 = root2 = -1;
	}

	public void header()
	{
		output("UNION/FIND Visualization\n", 4);
		output("Mr. Cris Kania <ckania@vt.edu>\n", 3);
		output("Dr. Cliff Shaffer <shaffer@vt.edu>\n", 3);
	}

	public void output(String str, int lvl)
	{
		try 
		{
			switch (lvl)
			{
				case 0: doc.insertString(doc.getLength(), str, textPane.getStyle("Style0")); break;
				case 1: doc.insertString(doc.getLength(), str, textPane.getStyle("Style1")); break;
				case 2: doc.insertString(doc.getLength(), "  " + str, textPane.getStyle("Style2")); break;
				case 3: doc.insertString(doc.getLength(), str, textPane.getStyle("Style3")); break;
				case 4: doc.insertString(doc.getLength(), str, textPane.getStyle("Style4")); break;
			}
		}
		catch (BadLocationException e)
		{
		}
	}
}
